//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/Multilayer/MultiLayer.cpp
//! @brief     Implements class MultiLayer.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/Multilayer/MultiLayer.h"
#include "Base/Math/IntegratorGK.h"
#include "Base/Util/Assert.h"
#include "Base/Util/StringUtil.h"
#include "Sample/Aggregate/ParticleLayout.h"
#include "Sample/Interface/LayerInterface.h"
#include "Sample/Interface/LayerRoughness.h"
#include "Sample/Material/MaterialUtil.h"
#include "Sample/Multilayer/Layer.h"
#include <numbers>

using std::numbers::pi;

MultiLayer::MultiLayer()
{
    validateOrThrow();
}

MultiLayer::~MultiLayer() = default;

MultiLayer* MultiLayer::clone() const
{
    auto* result = new MultiLayer;
    result->setExternalField(externalField());
    for (size_t i = 0; i < numberOfLayers(); ++i) {
        const LayerInterface* interface = i > 0 ? m_interfaces[i - 1] : nullptr;
        if (i > 0 && interface->roughness())
            result->addLayerWithTopRoughness(std::as_const(*m_layers[i]), *interface->roughness());
        else
            result->addLayer(std::as_const(*m_layers[i]));
    }
    return result;
}

void MultiLayer::checkMaterials(double wavelength) const
{
    for (size_t i = 0; i < numberOfLayers(); ++i)
        m_layers[i]->material()->checkRefractiveIndex(wavelength);
}

//! Adds layer with default (zero) roughness
void MultiLayer::addLayer(const Layer& layer)
{
    addLayerExec(layer, nullptr);
}

//! Adds layer with top roughness
void MultiLayer::addLayerWithTopRoughness(const Layer& layer, const LayerRoughness& roughness)
{
    addLayerExec(layer, &roughness);
}

//! Internal
void MultiLayer::addLayerExec(const Layer& layer, const LayerRoughness* roughness)
{
    const Layer* new_layer = layer.clone();

    if (numberOfLayers()) { // not the top layer
        const Layer* last_layer = m_layers.back();
        const LayerRoughness* new_roughness;
        if (roughness)
            new_roughness = roughness->clone();
        else {
            K_CorrelationModel autocorrelation(0, 0, 0);
            ErfInterlayer interlayer;
            new_roughness = new LayerRoughness(&autocorrelation, &interlayer);
        }
        m_interfaces.push_back(new LayerInterface(last_layer, new_layer, new_roughness));

    } else { // the top layer
        if (new_layer->thickness() != 0.0)
            throw std::runtime_error("Invalid top layer; to indicate that it is semiinfinite,"
                                     " it must have a nominal thickness of 0");
        if (roughness)
            throw std::runtime_error("Invalid top layer with roughness");
    }

    m_layers.push_back(new_layer);
    m_validated = false;
}

const AutocorrelationModel* MultiLayer::autocorrAt(int i_interface) const
{
    return m_interfaces.at(i_interface)->roughness()->autocorrelationModel();
}

double MultiLayer::maxCutoffSpatialFrequencyAt(size_t i_interface) const
{
    double result = 0;
    for (size_t i = i_interface; i < m_interfaces.size(); i++) {
        const auto autocorr = m_interfaces.at(i_interface)->roughness()->autocorrelationModel();
        result = std::max(autocorr->maxSpatialFrequency(), result);
    }
    return result;
}

const Layer* MultiLayer::layer(size_t i_layer) const
{
    return m_layers.at(i_layer);
}

const LayerInterface* MultiLayer::layerInterface(size_t i_interface) const
{
    return m_interfaces.at(i_interface);
}

void MultiLayer::setExternalField(const R3& ext_field)
{
    m_ext_field = ext_field;
    m_validated = false;
}

double MultiLayer::layerInterfaceSpectrum(double spatial_f, int i_interface) const
{
    int size = m_interfaces.size();
    const AutocorrelationModel* autocorr = autocorrAt(i_interface);

    if (const auto k_corr = dynamic_cast<const K_CorrelationModel*>(autocorr))
        return k_corr->spectralFunction(spatial_f);

    if (dynamic_cast<const LinearGrowthModel*>(autocorr)) {
        if (i_interface == size - 1)
            ASSERT_NEVER;

        int j = i_interface + 1;
        for (; j < size; j++) {
            if (dynamic_cast<const K_CorrelationModel*>(autocorrAt(j)))
                break;
        }
        ASSERT(j < size);
        const auto base_k_corr = dynamic_cast<const K_CorrelationModel*>(autocorrAt(j));
        double spectrum_below = base_k_corr->spectralFunction(spatial_f);
        double current_spectrum = spectrum_below;

        // all interfaces between i and j have linear growth model
        for (int k = j - 1; k >= i_interface; k--) {
            const auto lin_growth = dynamic_cast<const LinearGrowthModel*>(autocorrAt(k));
            const double thickness = m_layers.at(k + 1)->thickness();
            current_spectrum = lin_growth->spectralFunction(spectrum_below, thickness, spatial_f);
            spectrum_below = current_spectrum;
        }
        return current_spectrum;
    }
    ASSERT_NEVER;
}

double MultiLayer::layerInterfaceRMS(size_t i_interface) const
{
    ASSERT(i_interface < m_interfaces.size());
    const auto autocorr = m_interfaces.at(i_interface)->roughness()->autocorrelationModel();

    if (auto* k_corr = dynamic_cast<const K_CorrelationModel*>(autocorr)) {
        return k_corr->rms();
    } else if (dynamic_cast<const LinearGrowthModel*>(autocorr)) {
        const double maxSpatialFrequency = maxCutoffSpatialFrequencyAt(i_interface);
        return std::sqrt((2 * pi)
                         * RealIntegrator().integrate(
                             [this, i_interface](double spatial_f) -> double {
                                 return spatial_f * layerInterfaceSpectrum(spatial_f, i_interface);
                             },
                             0, maxSpatialFrequency));
    }
    ASSERT_NEVER;
}

std::vector<const INode*> MultiLayer::nodeChildren() const
{
    std::vector<const INode*> result;
    const size_t N = m_layers.size();
    result.reserve(N + m_interfaces.size());

    for (size_t i = 0; i < N; ++i) {
        result.push_back(layer(i));
        if (i == N - 1)
            break;
        result.push_back(layerInterface(i));
    }
    return result;
}

double MultiLayer::hig(size_t i) const
{
    ASSERT(m_validated);
    ASSERT(i >= 1 && i < numberOfLayers());
    return ZInterfaces.at(i - 1);
}

double MultiLayer::low(size_t i) const
{
    ASSERT(m_validated);
    ASSERT(i < numberOfLayers() - 1);
    return ZInterfaces.at(i);
}

std::string MultiLayer::validate() const
{
    std::vector<std::string> errs;
    if (MaterialUtil::checkMaterialTypes(containedMaterials())
        == MATERIAL_TYPES::InvalidMaterialType)
        errs.push_back("Sample contains incompatible material definitions");

    for (size_t i = 0; i < m_layers.size(); ++i) {
        std::string err = m_layers[i]->validate();
        if (!err.empty())
            errs.push_back("{ layer " + std::to_string(i) + ": " + err + " }");
    }

    for (size_t i = 0; i < m_interfaces.size(); ++i) {
        std::string err = m_interfaces[i]->validate();
        if (!err.empty())
            errs.push_back("{ interface " + std::to_string(i) + ": " + err + " }");
    }

    if (!errs.empty())
        return "[ " + Base::String::join(errs, ", ") + " ]";

    //! Precompute interface coordinates
    size_t N = numberOfLayers();
    if (N >= 1)
        ZInterfaces.resize(N - 1);
    if (N >= 2) {
        ZInterfaces[0] = 0;
        for (size_t i = 1; i < N - 1; ++i)
            ZInterfaces.at(i) = ZInterfaces.at(i - 1) - m_layers.at(i)->thickness();
    }

    m_validated = true;
    return "";
}
